/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "snes9x.h"
#include "memmap.h"
#include "debug.h"
#include "cpuexec.h"
#include "ppu.h"
#include "snapshot.h"
#include "apu.h"
#include "display.h"
#include "gfx.h"
#include "soundux.h"
#include "spc700.h"
#include "spc7110.h"
#include "psp.h"

#include "pg.h"
#include "filer.h"

#define RELEASE

static struct timeval	s_tvStart;
static int				s_iFrame;

char					PBPPath[MAX_PATH];
char					RomPath[MAX_PATH];
#define ANALOG_NONE		0
#define ANALOG_LEFT		1
#define ANALOG_RIGHT	2
#define ANALOG_UP		3
#define ANALOG_DOWN		4
static int 				s_iAnalog = ANALOG_NONE;
static int 				s_iOld_Analog = ANALOG_NONE;
enum {
	AKEY_NONE = 0,
	AKEY_FRAMESKIPUP,
	AKEY_FRAMESKIPDOWN,
	AKEY_AUTOSKIP,
	AKEY_SCRSIZE,
	AKEY_SHOWFPS,
	AKEY_VSYNC,
	AKEY_SOUNDOFF,
	AKEY_TRANS,
};

extern "C" { void save_config(void); };

#define PSP_CFG_VERSION "uo_Snes9x for PSP Ver.0.02y28n1"

typedef struct
{
	char	vercnf[48];
	int		iSaveSlot;
	int		iSkipFrames;
	bool8	bShowFPS;
	bool8	bVSync;
	bool8	bSoundOff;
	int		iSoundRate;
	bool8	bTrans;
	int		iHBlankCycleDiv;
	int		iAPUTimerCycleDiv;
	bool8	bSwapAnalog;
	bool8	bSaveThumb;
	int		iPSP_ClockUp;
	int		iScreenSize;
	bool8	bAutoSkip;
	int		iKeyConfig[8];
	int		iAKeyConfig[3];
	bool8	bAkeyLock;
	bool8	bWPad;
} PSPSETTINGS;
PSPSETTINGS PSP_Settings;

typedef struct
{
	char 		LastPath[MAX_PATH];
	PSPSETTINGS	Default;
} GLOBALSETTINGS;
GLOBALSETTINGS Global_Settings;

#define UPPER_THRESHOLD		0xcf
#define LOWER_THRESHOLD		0x2f
#define PSP_CYCLE_DIV_MAX	30

volatile bool8			g_bSleep = false;
volatile bool8			g_bROMLoaded = false;

volatile bool8			g_bLoop = true;
int						g_thread = -1;
static uint8			SoundBuffer[MAX_BUFFER_SIZE];

#define SOUND_SAMPLE	1024
#define MAXVOLUME		0x8000

#define timercmp(a, b, CMP)	(((a)->tv_sec == (b)->tv_sec) ? ((a)->tv_usec CMP (b)->tv_usec) : ((a)->tv_sec CMP (b)->tv_sec))

void S9xCloseSoundDevice();

#include "psp2.cpp"

extern "C" {
void *S9xProcessSound (void *);

#ifndef RELEASE
int format_int( char* buf, int value )
{
	char*	org;
	int		div;
	int		val;
	bool	bFirst;
	int		i;

	org    = buf;
	bFirst = true;
	div    = 1000000000;
	for ( i = 0; i < 10; i++ ){
		val = (unsigned)value / div;

		if ( !bFirst || val ){
			*buf++ = val + '0';
			bFirst = false;
		}

		value %= div;
		div   /= 10;
	}

	if ( bFirst ){
		*buf++ = '0';
	}
	*buf = 0;

	return strlen( org );
}
#endif // RELEASE

void debug_log( const char* message )
{
#ifndef RELEASE
	static int	sy = 1;

	if ( sy >= CMAX_Y ) {
		pgFillBoxWF(256, 0, 479, 271, 0);
		sy = 1;
	}
	pgPrintWF( SNES_WIDTH / 8, sy, 0xffff, message );
	sy++;
#endif // RELEASE
}

void debug_int( const char* message, int value )
{
#ifndef RELEASE
	strcpy( String, message );
	format_int( &String[strlen( String )], value );

	debug_log( String );
#endif // RELEASE
}

void debug_hex( int value )
{
#ifndef RELEASE
	int		shift;
	int		val;
	int		i;

	shift = 28;
	for ( i = 0; i < 8; i++ ){
		val = (value >> shift) & 0x0f;
		if ( val < 10 ){
			String[i] = val + '0';
		} else {
			String[i] = val - 10 + 'A';
		}
		shift -= 4;
	}
	String[i] = 0;

	debug_log( String );
#endif // RELEASE
}

static struct timeval s_analyze;
void StartAnalyze()
{
	sceKernelLibcGettimeofday(&s_analyze, NULL);
}

void StopAnalyze()
{
	struct timeval now;
	int		diff;

	sceKernelLibcGettimeofday(&now, NULL);

	diff  = (now.tv_sec - s_analyze.tv_sec) * 1000000 + now.tv_usec - s_analyze.tv_usec;

	debug_int( "time:", diff );
}

}; // end of extern "C"

//
// C++ Language
//
#ifndef OPTI
void JustifierButtons(uint32&)
{
	
}

bool JustifierOffscreen()
{
	return false;
}

void S9xAutoSaveSRAM()
{
//	Memory.SaveSRAM (S9xGetFilename("srm"));
}
#endif // OPTI

bool8 S9xOpenSoundDevice( int mode, bool8 stereo, int buffer_size )
{
	so.mute_sound  = TRUE;

	if ( buffer_size <= 0 ){
		return FALSE;
	}

#ifndef OPTI
	so.sound_switch = 255;
	so.stereo       = TRUE;
	so.sixteen_bit  = TRUE;
#endif // OPTI
	so.buffer_size  = buffer_size;
	so.encoded      = FALSE;

	// Initialize channel and allocate buffer
	so.sound_fd = sceAudioChReserve(-1, buffer_size, 0);
	if ( so.sound_fd < 0 ){
		return FALSE;
	}

#ifdef OPTI
	so.buffer_size <<= 2;
#else
	if ( so.stereo ){
		so.buffer_size *= 2;
	}
	if ( so.sixteen_bit ){
		so.buffer_size *= 2;
	}
#endif // OPTI
	if ( so.buffer_size > MAX_BUFFER_SIZE ){
		so.buffer_size = MAX_BUFFER_SIZE;
	}

	S9xSetPlaybackRate( 44100 );

	so.mute_sound  = FALSE;

	return TRUE;
}

void S9xCloseSoundDevice()
{
	if ( so.sound_fd >= 0 ){
		sceAudioChRelease(so.sound_fd);
		so.sound_fd = -1;
	}
}

bool8 S9xOpenSnapshotFile( const char *fname, bool8 read_only, STREAM *file)
{
    char filename [_MAX_PATH + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fn [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];

    _splitpath( fname, drive, dir, fn, ext);
    _makepath( filename, drive, dir, fn, ext);

    if (read_only)
    {
	if ((*file = OPEN_STREAM (filename, "rb")))
	    return (TRUE);
    }
    else
    {
	if ((*file = OPEN_STREAM (filename, "wb")))
	    return (TRUE);
        FILE *fs = fopen (filename, "rb");
        if (fs)
        {
            sprintf (String, "kt@C݂͑Ă܂݋֎~łB");
            fclose (fs);
            S9xMessage (S9X_ERROR, S9X_FREEZE_FILE_NOT_FOUND, String);
        }
        else
        {
            sprintf (String, "kt@C쐬邱Ƃł܂B fBNg͏݋֎~݂Ă܂B");
            S9xMessage (S9X_ERROR, S9X_FREEZE_FILE_NOT_FOUND, String);
        }
    }
    return (FALSE);
}

void S9xCloseSnapshotFile( STREAM file)
{
    CLOSE_STREAM (file);
}

//
// C Language
//
extern "C" {
void S9xMessage( int type, int number, const char* message )
{
#ifndef RELEASE
	debug_log( message );
#endif // RELEASE
	S9xSetInfoString( message );
}

void S9xMovieUpdate ()
{
	
}


bool8 S9xReadSuperScopePosition (int &x, int &y, uint32 &buttons)
{
	return FALSE;
}

bool8 S9xReadMousePosition (int which, int &x, int &y, uint32 &buttons)
{
	return FALSE;
}

void S9xLoadSDD1Data()
{
	Settings.SDD1Pack=FALSE;
	Memory.FreeSDD1Data ();
	
	if (strncmp (Memory.ROMName, "Star Ocean", 10) == 0) {
		Settings.SDD1Pack=TRUE;
	} else if(strncmp (Memory.ROMName, "STREET FIGHTER ALPHA2", 21)==0) {
		Settings.SDD1Pack=TRUE;
	} else if(strncmp (Memory.ROMName, "STREET FIGHTER ZERO2", 20)==0) {
		Settings.SDD1Pack=TRUE;
	}
}

uint32 S9xReadJoypad( int which1 )
{
	if ( which1 ){
		return 0;
	}
	
	ctrl_data_t ctl;
	uint32 ret;
	
	sceCtrlPeekBufferPositive(&ctl, 1);
	
	ret = 0x80000000;
	
	if (ctl.buttons & PSP_Settings.iKeyConfig[0]) ret |= SNES_X_MASK;
	if (ctl.buttons & PSP_Settings.iKeyConfig[1]) ret |= SNES_Y_MASK;
	if (ctl.buttons & PSP_Settings.iKeyConfig[2]) ret |= SNES_A_MASK;
	if (ctl.buttons & PSP_Settings.iKeyConfig[3]) ret |= SNES_B_MASK;
	if (ctl.buttons & PSP_Settings.iKeyConfig[4]) ret |= SNES_TL_MASK;
	if (ctl.buttons & PSP_Settings.iKeyConfig[5]) ret |= SNES_TR_MASK;
	if (ctl.buttons & PSP_Settings.iKeyConfig[6]) ret |= SNES_START_MASK;
	if (ctl.buttons & PSP_Settings.iKeyConfig[7]) ret |= SNES_SELECT_MASK;
	
	if (PSP_Settings.bSwapAnalog) {
		if (!PSP_Settings.bAkeyLock || (ctl.buttons & CTRL_SELECT)) {
			if (ctl.analogX < LOWER_THRESHOLD) ret |= SNES_LEFT_MASK;
			if (ctl.analogX > UPPER_THRESHOLD) ret |= SNES_RIGHT_MASK;
			if (ctl.analogY < LOWER_THRESHOLD) ret |= SNES_UP_MASK;
			if (ctl.analogY > UPPER_THRESHOLD) ret |= SNES_DOWN_MASK;
			s_iAnalog = ANALOG_NONE;
			if ( ctl.buttons & CTRL_LEFT ){
				s_iAnalog = ANALOG_LEFT;
			} else if ( ctl.buttons & CTRL_RIGHT ){
				s_iAnalog = ANALOG_RIGHT;
			} else if ( ctl.buttons & CTRL_UP ){
				s_iAnalog = ANALOG_UP;
			} else if ( ctl.buttons & CTRL_DOWN ){
				s_iAnalog = ANALOG_DOWN;
			}
		} else {
			if (ctl.analogX < LOWER_THRESHOLD) ret |= SNES_LEFT_MASK;
			if (ctl.analogX > UPPER_THRESHOLD) ret |= SNES_RIGHT_MASK;
			if (ctl.analogY < LOWER_THRESHOLD) ret |= SNES_UP_MASK;
			if (ctl.analogY > UPPER_THRESHOLD) ret |= SNES_DOWN_MASK;
			if (PSP_Settings.bWPad) {
				if (ctl.buttons & CTRL_UP) ret |= SNES_UP_MASK;
				if (ctl.buttons & CTRL_DOWN) ret |= SNES_DOWN_MASK;
				if (ctl.buttons & CTRL_RIGHT) ret |= SNES_RIGHT_MASK;
				if (ctl.buttons & CTRL_LEFT) ret |= SNES_LEFT_MASK;
			}
			s_iAnalog = ANALOG_NONE;
		}
	} else {
		if (!PSP_Settings.bAkeyLock || (ctl.buttons & CTRL_SELECT)) {
			if (ctl.buttons & CTRL_UP) ret |= SNES_UP_MASK;
			if (ctl.buttons & CTRL_DOWN) ret |= SNES_DOWN_MASK;
			if (ctl.buttons & CTRL_RIGHT) ret |= SNES_RIGHT_MASK;
			if (ctl.buttons & CTRL_LEFT) ret |= SNES_LEFT_MASK;
			s_iAnalog = ANALOG_NONE;
			if ( ctl.analogX == 0x00 ){
				s_iAnalog = ANALOG_LEFT;
			} else if ( ctl.analogX == 0xff ){
				s_iAnalog = ANALOG_RIGHT;
			} else if ( ctl.analogY == 0x00 ){
				s_iAnalog = ANALOG_UP;
			} else if ( ctl.analogY == 0xff ){
				s_iAnalog = ANALOG_DOWN;
			}
		} else {
			if (ctl.buttons & CTRL_UP) ret |= SNES_UP_MASK;
			if (ctl.buttons & CTRL_DOWN) ret |= SNES_DOWN_MASK;
			if (ctl.buttons & CTRL_RIGHT) ret |= SNES_RIGHT_MASK;
			if (ctl.buttons & CTRL_LEFT) ret |= SNES_LEFT_MASK;
			if (PSP_Settings.bWPad) {
				if (ctl.analogX < LOWER_THRESHOLD) ret |= SNES_LEFT_MASK;
				if (ctl.analogX > UPPER_THRESHOLD) ret |= SNES_RIGHT_MASK;
				if (ctl.analogY < LOWER_THRESHOLD) ret |= SNES_UP_MASK;
				if (ctl.analogY > UPPER_THRESHOLD) ret |= SNES_DOWN_MASK;
			}
			s_iAnalog = ANALOG_NONE;
		}
	}
	return ret;
}

void S9xSetPalette()
{
	
}

void S9xGenerateSound()
{
	
}

void S9xExit()
{
	Memory.Deinit();
}

void *S9xProcessSound (void *)
{
	do {
#ifdef OPTI
		unsigned byte_offset;
		unsigned bytes_to_write;
		byte_offset = so.play_position;
		S9xMixSamples (SoundBuffer + byte_offset, (SOUND_SAMPLE << 1));
		bytes_to_write = (SOUND_SAMPLE << 2);
		so.play_position = (so.play_position + (SOUND_SAMPLE << 2)) & SOUND_BUFFER_SIZE_MASK;
#else
		int sample_count;
		unsigned byte_offset;
		unsigned bytes_to_write;
		sample_count = so.buffer_size >> 1;
		byte_offset = so.play_position;
		S9xMixSamples (SoundBuffer + byte_offset, sample_count);
		bytes_to_write = so.buffer_size;
		so.play_position = (so.play_position + so.buffer_size) & SOUND_BUFFER_SIZE_MASK;
#endif // OPTI
		for(;;) {
			int I = bytes_to_write;
			if (byte_offset + I > SOUND_BUFFER_SIZE) {
				I = SOUND_BUFFER_SIZE - byte_offset;
			}
			if(I == 0) break;
			sceAudioOutputPannedBlocking(so.sound_fd, MAXVOLUME, MAXVOLUME, (char*)SoundBuffer + byte_offset);
			if (I > 0) {
				bytes_to_write -= I;
				byte_offset += I;
				byte_offset &= SOUND_BUFFER_SIZE_MASK; /* wrap */
			}
		}
	} while (Settings.ThreadSound);
	return (NULL);
}

void InitTimer()
{
	debug_log( "Create Thread" );
	g_thread = sceKernelCreateThread( "sound thread", (threadfunc_t)S9xProcessSound, 0x12, 0x40000, 0, NULL );
	if ( g_thread < 0 ){
		debug_log( "Thread failed" );
		return;
	}

	sceKernelStartThread( g_thread, 0, 0 );

	debug_log( "Thread ok" );
}

static struct timeval next1 = {0, 0};
void S9xSyncSpeed()
{
#ifndef RELEASE
	S9xProcessEvents( FALSE );

	IPPU.FrameSkip = 0;
	IPPU.SkippedFrames = 0;
	IPPU.RenderThisFrame = TRUE;

#else
	S9xProcessEvents( FALSE );

	if (!PSP_Settings.bAutoSkip) {
		if (++IPPU.FrameSkip >= Settings.SkipFrames) {
			IPPU.FrameSkip = 0;
			IPPU.SkippedFrames = 0;
			IPPU.RenderThisFrame = TRUE;
		} else {
			IPPU.SkippedFrames++;
			IPPU.RenderThisFrame = FALSE;
		}
	} else {
		struct timeval now;
		
		sceKernelLibcGettimeofday(&now, NULL);
		if (next1.tv_sec == 0) {
			next1 = now;
			next1.tv_usec++;
		}
		if (timercmp(&next1, &now, >)) {
			if (IPPU.SkippedFrames == 0) {
				unsigned int timeleft = (next1.tv_sec - now.tv_sec) * 1000000
									+ next1.tv_usec - now.tv_usec;
				if (timeleft < Settings.FrameTime)
					sceKernelDelayThread(timeleft);
				sceKernelLibcGettimeofday(&now, NULL);
				next1 = now;
			}
			IPPU.SkippedFrames = 0;
			IPPU.RenderThisFrame = TRUE;
		} else {
			if (IPPU.SkippedFrames < Settings.SkipFrames) {
				IPPU.SkippedFrames++;
				IPPU.RenderThisFrame = FALSE;
			} else {
				IPPU.SkippedFrames = 0;
				IPPU.RenderThisFrame = TRUE;
				next1 = now;
			}
		}
		next1.tv_usec += Settings.FrameTime;
		if (next1.tv_usec >= 1000000) {
			next1.tv_sec += next1.tv_usec / 1000000;
			next1.tv_usec %= 1000000;
		}
	}
#endif // RELEASE
}

const char *S9xGetFilename( const char *e )
{
	static char filename [_MAX_PATH + 1];
	char drive [_MAX_DRIVE + 1];
	char dir [_MAX_DIR + 1];
	char fname [_MAX_FNAME + 1];
	char ext [_MAX_EXT + 1];

	_splitpath (Memory.ROMFilename, drive, dir, fname, ext);

	if (strcasecmp (e, "srm") == 0) {
		strcpy(filename,PBPPath);
		strcat(filename,"SAVE/");
		strcat(filename,fname);
		strcat(filename,".srm");
	} else if (strcasecmp (e, "sv0") == 0) {
		strcpy(filename,PBPPath);
		strcat(filename,"SAVE/");
		strcat(filename,fname);
		strcat(filename,".000");
		filename[strlen(filename)-1] = PSP_Settings.iSaveSlot + '0';
	} else if (strcasecmp (e, "tn0") == 0) {
		strcpy(filename,PBPPath);
		strcat(filename,"SAVE/");
		strcat(filename,fname);
		strcat(filename,".tn0");
		filename[strlen(filename)-1] = PSP_Settings.iSaveSlot + '0';
	} else if (strcasecmp (e, "cfg") == 0) {
		strcpy(filename,PBPPath);
		strcat(filename,"SAVE/");
		strcat(filename,fname);
		strcat(filename,"_psp.cfg");
	} else {
		_makepath (filename, drive, dir, fname, e);
	}
	return (filename);
}

bool8 S9xInitUpdate()
{
	return TRUE;
}

enum {
	SCR_X1 = 0,
	SCR_FIT,
	SCR_43FIT,
	SCR_WX2,
	SCR_FULLFIT,
	SCR_GEFIT,
	SCR_GE43FIT,
	SCR_GEFULLFIT,
	SCR_MAX = SCR_GEFULLFIT,
};

static uint8 *GFX_Screen0 = (uint8 *)0x040CC000;
static uint8 *GFX_Screen1 = (uint8 *)0x0414C000;
static uint8 GFX_SubScreen[512 * 512 * 2];
static uint8 GFX_ZBuffer[512 * 512 * 2];
static uint8 GFX_SubZBuffer[512 * 512 * 2];

static int GeFlip = 0;
static unsigned int *GEcmd = (unsigned int *)0x441CC000;
static short *ScreenVertex = (short *)0x441CC100;

static int gecbid = -1;

void PSP_ShowFPS(void)
{
	struct timeval now;
	unsigned int diff, fps;
	static char FPSbuf[7];
	
	s_iFrame++;
	sceKernelLibcGettimeofday(&now, NULL);
	
	diff  = (now.tv_sec - s_tvStart.tv_sec) * 1000000 + now.tv_usec - s_tvStart.tv_usec;
	diff /= 1000000;
	
	if ( diff ){
		fps = s_iFrame / diff;
		if (fps >=100) {
			FPSbuf[0] = (fps / 100) + '0';
			fps -= (fps / 100) * 100;
		} else {
			FPSbuf[0] = ' ';
		}
		FPSbuf[1] = (fps / 10) + '0';
		FPSbuf[2] = (fps % 10) + '0';
		FPSbuf[3] = 'F';
		FPSbuf[4] = 'P';
		FPSbuf[5] = 'S';
		FPSbuf[6] = '\0';
	
		s_tvStart = now;
		s_iFrame  = 0;
	}
	pgPrintBG( CMAX_X - 6, 0, 0xffff, FPSbuf );
}

void Ge_Finish_Callback(int id, void *arg)
{
	if (PSP_Settings.bShowFPS) PSP_ShowFPS();
	if (PSP_Settings.bVSync) sceDisplayWaitVblank();
	pgScreenFlip();
}

#ifdef OPTI
bool8 S9xDeinitUpdate (int Width, int Height)
#else
bool8 S9xDeinitUpdate (int Width, int Height, bool8 sixteen_bit)
#endif // OPTI
{
#ifdef RELEASE
	static int qid = -1;
	unsigned int log2w, log2h;

	switch(PSP_Settings.iScreenSize) {
	default:
	case SCR_X1:
		break;
	case SCR_FIT:
		pgBitBltFit((unsigned short *)GFX.Screen, Height);
		break;
	case SCR_43FIT:
		pgBitBlt43Fit((unsigned short *)GFX.Screen, Height);
		break;
	case SCR_WX2:
		pgBitBltWx2((unsigned long *)GFX.Screen, Height);
		break;
	case SCR_FULLFIT:
		pgBitBltFullFit((unsigned short *)GFX.Screen, Height);
		break;
	case SCR_GEFIT:
		if (qid >= 0) sceGeListSync(qid, 0);
		ScreenVertex[2] = 84;
		ScreenVertex[5] = Width;
		ScreenVertex[6] = Height;
		ScreenVertex[7] = 84 + 312;
		GEcmd[ 0] = 0x9C000000UL | ((uint32)(unsigned char *)pgGetVramAddr( 0, 0 ) & 0x00FFFFFF);
		if (GeFlip == 0) {
			GEcmd[ 2] = 0xA0000000UL | ((uint32)(unsigned char *)GFX_Screen0 & 0x00FFFFFF);
			GFX.Screen = (uint8*)GFX_Screen1;
			GeFlip = 1;
		} else {
			GEcmd[ 2] = 0xA0000000UL | ((uint32)(unsigned char *)GFX_Screen1 & 0x00FFFFFF);
			GFX.Screen = (uint8*)GFX_Screen0;
			GeFlip = 0;
		}
		if (Width > 256) log2w = 9;
		else log2w = 8;
		if (Height > 256) log2h = 9;
		else log2h = 8;
		GEcmd[ 4] = 0xB8000000UL | (log2h << 8) | log2w;
		sceKernelDcacheWritebackAll();
		qid = sceGeListEnQueue(&GEcmd[0], &GEcmd[14], gecbid, NULL);
		return TRUE;
	case SCR_GE43FIT:
		if (qid >= 0) sceGeListSync(qid, 0);
		ScreenVertex[2] = 59;
		ScreenVertex[5] = Width;
		ScreenVertex[6] = Height;
		ScreenVertex[7] = 59 + 362;
		GEcmd[ 0] = 0x9C000000UL | ((uint32)(unsigned char *)pgGetVramAddr( 0, 0 ) & 0x00FFFFFF);
		if (GeFlip == 0) {
			GEcmd[ 2] = 0xA0000000UL | ((uint32)(unsigned char *)GFX_Screen0 & 0x00FFFFFF);
			GFX.Screen = (uint8*)GFX_Screen1;
			GeFlip = 1;
		} else {
			GEcmd[ 2] = 0xA0000000UL | ((uint32)(unsigned char *)GFX_Screen1 & 0x00FFFFFF);
			GFX.Screen = (uint8*)GFX_Screen0;
			GeFlip = 0;
		}
		if (Width > 256) log2w = 9;
		else log2w = 8;
		if (Height > 256) log2h = 9;
		else log2h = 8;
		GEcmd[ 4] = 0xB8000000UL | (log2h << 8) | log2w;
		sceKernelDcacheWritebackAll();
		qid = sceGeListEnQueue(&GEcmd[0], &GEcmd[14], gecbid, NULL);
		return TRUE;
	case SCR_GEFULLFIT:
		if (qid >= 0) sceGeListSync(qid, 0);
		ScreenVertex[2] = 0;
		ScreenVertex[5] = Width;
		ScreenVertex[6] = Height;
		ScreenVertex[7] = 480;
		GEcmd[ 0] = 0x9C000000UL | ((uint32)(unsigned char *)pgGetVramAddr( 0, 0 ) & 0x00FFFFFF);
		if (GeFlip == 0) {
			GEcmd[ 2] = 0xA0000000UL | ((uint32)(unsigned char *)GFX_Screen0 & 0x00FFFFFF);
			GFX.Screen = (uint8*)GFX_Screen1;
			GeFlip = 1;
		} else {
			GEcmd[ 2] = 0xA0000000UL | ((uint32)(unsigned char *)GFX_Screen1 & 0x00FFFFFF);
			GFX.Screen = (uint8*)GFX_Screen0;
			GeFlip = 0;
		}
		if (Width > 256) log2w = 9;
		else log2w = 8;
		if (Height > 256) log2h = 9;
		else log2h = 8;
		GEcmd[ 4] = 0xB8000000UL | (log2h << 8) | log2w;
		sceKernelDcacheWritebackAll();
		qid = sceGeListEnQueue(&GEcmd[0], &GEcmd[14], gecbid, NULL);
		return TRUE;
	}
#endif // RELEASE
	if (PSP_Settings.bShowFPS) PSP_ShowFPS();
	if (PSP_Settings.bVSync) sceDisplayWaitVblank();
	pgScreenFlip();

#ifdef RELEASE
	if (PSP_Settings.iScreenSize == SCR_X1)
		GFX.Screen = (uint8*)pgGetVramAddr( (SCREEN_WIDTH - SNES_WIDTH) >> 1, (SCREEN_HEIGHT - SNES_HEIGHT) >> 1 );
#else
	GFX.Screen = (uint8*)pgGetVramAddr( 0, 0 );
#endif // RELEASE
	return TRUE;
}

void S9xInitDisplay( int argc, char** argv )
{
	PSP_Settings.bTrans = true;
	Settings.Transparency = TRUE;
#ifndef OPTI
	Settings.SixteenBit   = TRUE;
#endif // OPTI
	Settings.SupportHiRes = FALSE; //interpolate;

	memset( GFX_Screen0,    0, 0x80000 );
	memset( GFX_Screen1,    0, 0x80000 );
	memset( GFX_SubScreen,  0, sizeof(GFX_SubScreen) );
	memset( GFX_ZBuffer,    0, sizeof(GFX_ZBuffer) );
	memset( GFX_SubZBuffer, 0, sizeof(GFX_SubZBuffer) );

	GFX.Pitch      = 512 * 2;
#ifdef RELEASE
	ScreenVertex[0] = 0;
	ScreenVertex[1] = 0;
	ScreenVertex[2] = 0;
	ScreenVertex[3] = 0;
	ScreenVertex[4] = 0;
	ScreenVertex[5] = SNES_WIDTH;
	ScreenVertex[6] = SNES_HEIGHT;
	ScreenVertex[7] = SCREEN_WIDTH;
	ScreenVertex[8] = SCREEN_HEIGHT;
	ScreenVertex[9] = 0;
	// Set Draw Buffer
	GEcmd[ 0] = 0x9C000000UL | ((uint32)(unsigned char *)pgGetVramAddr( 0, 0 ) & 0x00FFFFFF);
	GEcmd[ 1] = 0x9D000000UL | (((uint32)(unsigned char *)pgGetVramAddr( 0, 0 ) & 0xFF000000) >> 8) | 512;
	// Set Tex Buffer
	GEcmd[ 2] = 0xA0000000UL | ((uint32)(unsigned char *)GFX_Screen0 & 0x00FFFFFF);
	GEcmd[ 3] = 0xA8000000UL | (((uint32)(unsigned char *)GFX_Screen0 & 0xFF000000) >> 8) | 512;
	// Tex size
	GEcmd[ 4] = 0xB8000000UL | (9 << 8) | 9;
	// Tex Flush
	GEcmd[ 5] = 0xCB000000UL;
	// Set Vertex
	GEcmd[ 6] = 0x12000000UL | (1 << 23) | (0 << 11) | (0 << 9) | (2 << 7) | (0 << 5) | (0 << 2) | 2;
	GEcmd[ 7] = 0x10000000UL;
	GEcmd[ 8] = 0x02000000UL;
	GEcmd[ 9] = 0x10000000UL | (((uint32)(void *)ScreenVertex & 0xFF000000) >> 8);
	GEcmd[10] = 0x01000000UL | ((uint32)(void *)ScreenVertex & 0x00FFFFFF);
	// Draw Vertex
	GEcmd[11] = 0x04000000UL | (6 << 16) | 2;
	// List End
	GEcmd[12] = 0x0F000000UL;
	GEcmd[13] = 0x0C000000UL;
	GEcmd[14] = 0;
	GEcmd[15] = 0;
	GeFlip = 0;

	GFX.Screen     = (uint8*)pgGetVramAddr( (SCREEN_WIDTH - SNES_WIDTH) >> 1, (SCREEN_HEIGHT - SNES_HEIGHT) >> 1 );
#else
	GFX.Screen     = (uint8*)pgGetVramAddr( 0, 0 );
#endif // RELEASE
	GFX.SubScreen  = (uint8*)GFX_SubScreen;
	GFX.ZBuffer    = (uint8*)GFX_ZBuffer;
	GFX.SubZBuffer = (uint8*)GFX_SubZBuffer;
}

void S9xInitInputDevices()
{
	
}

void fix_H_Max(void)
{
	//CPU timing hacks
	// A Couple of HDMA related hacks - Lantus
	if ((strcmp(Memory.ROMName, "SFX SUPERBUTOUDEN2")==0) ||
	    (strcmp(Memory.ROMName, "ALIEN vs. PREDATOR")==0) ||
		(strcmp(Memory.ROMName, "STONE PROTECTORS")==0) ||
	    (strcmp(Memory.ROMName, "SUPER BATTLETANK 2")==0))
		Settings.H_Max = (Settings.H_Max * 130) / 100;
	if(strcmp(Memory.ROMName, "HOME IMPROVEMENT")==0)
		Settings.H_Max = (Settings.H_Max * 200) / 100;
    if (strcmp (Memory.ROMId, "ASRJ") == 0 && Settings.CyclesPercentage == 100)
		// Street Racer
		Settings.H_Max = (Settings.H_Max * 95) / 100;
	// Power Rangers Fight
    if (strncmp (Memory.ROMId, "A3R", 3) == 0 ||
        // Clock Tower
		strncmp (Memory.ROMId, "AJE", 3) == 0)
		Settings.H_Max = (Settings.H_Max * 103) / 100;
    if (strncmp (Memory.ROMId, "A3M", 3) == 0 && Settings.CyclesPercentage == 100)
		// Mortal Kombat 3. Fixes cut off speech sample
		Settings.H_Max = (Settings.H_Max * 110) / 100;
    if (strcmp (Memory.ROMName, "\x0bd\x0da\x0b2\x0d4\x0b0\x0bd\x0de") == 0 &&
		Settings.CyclesPercentage == 100)
		Settings.H_Max = (Settings.H_Max * 101) / 100;
    // Start Trek: Deep Sleep 9
    if (strncmp (Memory.ROMId, "A9D", 3) == 0 && Settings.CyclesPercentage == 100)
		Settings.H_Max = (Settings.H_Max * 110) / 100;
}

void key_config_init(void)
{
	PSP_Settings.iKeyConfig[0]=CTRL_TRIANGLE;		// X
	PSP_Settings.iKeyConfig[1]=CTRL_SQUARE;			// Y
	PSP_Settings.iKeyConfig[2]=CTRL_CIRCLE;			// A
	PSP_Settings.iKeyConfig[3]=CTRL_CROSS;			// B
	PSP_Settings.iKeyConfig[4]=CTRL_LTRIGGER;		// L
	PSP_Settings.iKeyConfig[5]=CTRL_RTRIGGER;		// R
	PSP_Settings.iKeyConfig[6]=CTRL_START;			// START
	PSP_Settings.iKeyConfig[7]=CTRL_SELECT;			// SELECT

	PSP_Settings.iAKeyConfig[0]=AKEY_AUTOSKIP;		// RIGHT
	PSP_Settings.iAKeyConfig[1]=AKEY_FRAMESKIPUP;	// UP
	PSP_Settings.iAKeyConfig[2]=AKEY_FRAMESKIPDOWN;	// DOWN

	PSP_Settings.bSwapAnalog = false;
	PSP_Settings.bAkeyLock = false;
	PSP_Settings.bWPad = false;
}

void key_config(void)
{
	enum {
		KEY_CONF_X = 0,
		KEY_CONF_Y,
		KEY_CONF_A,
		KEY_CONF_B,
		KEY_CONF_TL,
		KEY_CONF_TR,
		KEY_CONF_START,
		KEY_CONF_SELECT,

		KEY_CONF_A_RIGHT,
		KEY_CONF_A_UP,
		KEY_CONF_A_DOWN,
		KEY_CONF_SWAPAD,
		KEY_CONF_SELLOCK,
		KEY_CONF_WPAD,

		KEY_CONF_INIT,
		KEY_CONF_EXIT,
	};
	
	char msg[256];
	int sel=KEY_CONF_EXIT, x, y, i;
	
	for(;;){
		readpad();
		if (new_pad & CTRL_CIRCLE) {
			if (sel >= KEY_CONF_X && sel <= KEY_CONF_SELECT) {
				PSP_Settings.iKeyConfig[sel] = CTRL_CIRCLE;
			} else if (sel == KEY_CONF_A_RIGHT) {
				PSP_Settings.iAKeyConfig[0]++;
				if (PSP_Settings.iAKeyConfig[0] > AKEY_TRANS) {
					PSP_Settings.iAKeyConfig[0] = AKEY_NONE;
				}
			} else if (sel == KEY_CONF_A_UP) {
				PSP_Settings.iAKeyConfig[1]++;
				if (PSP_Settings.iAKeyConfig[1] > AKEY_TRANS) {
					PSP_Settings.iAKeyConfig[1] = AKEY_NONE;
				}
			} else if (sel == KEY_CONF_A_DOWN) {
				PSP_Settings.iAKeyConfig[2]++;
				if (PSP_Settings.iAKeyConfig[2] > AKEY_TRANS) {
					PSP_Settings.iAKeyConfig[2] = AKEY_NONE;
				}
			} else if (sel == KEY_CONF_SWAPAD) {
				PSP_Settings.bSwapAnalog = !PSP_Settings.bSwapAnalog;
			} else if (sel == KEY_CONF_SELLOCK) {
				PSP_Settings.bAkeyLock = !PSP_Settings.bAkeyLock;
			} else if (sel == KEY_CONF_WPAD) {
				PSP_Settings.bWPad = !PSP_Settings.bWPad;
			} else if (sel == KEY_CONF_INIT) {
				key_config_init();
			} else if (sel == KEY_CONF_EXIT) {
				break;
			}
		} else if (new_pad & CTRL_CROSS) {
			if (sel >= KEY_CONF_X && sel <= KEY_CONF_SELECT)
				PSP_Settings.iKeyConfig[sel] = CTRL_CROSS;
			else if (sel >= KEY_CONF_A_RIGHT && new_pad & CTRL_CROSS)
				break;
		} else if (new_pad & CTRL_TRIANGLE) {
			if (sel >= KEY_CONF_X && sel <= KEY_CONF_SELECT)
				PSP_Settings.iKeyConfig[sel] = CTRL_TRIANGLE;
		} else if (new_pad & CTRL_SQUARE) {
			if (sel >= KEY_CONF_X && sel <= KEY_CONF_SELECT)
				PSP_Settings.iKeyConfig[sel] = CTRL_SQUARE;
		} else if (new_pad & CTRL_LTRIGGER) {
			if (sel >= KEY_CONF_X && sel <= KEY_CONF_SELECT)
				PSP_Settings.iKeyConfig[sel] = CTRL_LTRIGGER;
		} else if (new_pad & CTRL_RTRIGGER) {
			if (sel >= KEY_CONF_X && sel <= KEY_CONF_SELECT)
				PSP_Settings.iKeyConfig[sel] = CTRL_RTRIGGER;
		} else if (new_pad & CTRL_START) {
			if (sel >= KEY_CONF_X && sel <= KEY_CONF_SELECT)
				PSP_Settings.iKeyConfig[sel] = CTRL_START;
		} else if (new_pad & CTRL_SELECT) {
			if (sel >= KEY_CONF_X && sel <= KEY_CONF_SELECT)
				PSP_Settings.iKeyConfig[sel] = CTRL_SELECT;
		} else if (new_pad & CTRL_UP){
			if (sel!=0){
				sel--;
			} else {
				sel=KEY_CONF_EXIT;
			}
		} else if (new_pad & CTRL_DOWN){
			if (sel!=KEY_CONF_EXIT){
				sel++;
			} else {
				sel=0;
			}
		} else if (new_pad & CTRL_LEFT){
			if (sel == KEY_CONF_A_RIGHT){
				if (PSP_Settings.iAKeyConfig[0] < 1) {
					PSP_Settings.iAKeyConfig[0] = 1;
				}
				PSP_Settings.iAKeyConfig[0]--;
			} else if (sel == KEY_CONF_A_UP){
				if (PSP_Settings.iAKeyConfig[1] < 1) {
					PSP_Settings.iAKeyConfig[1] = 1;
				}
				PSP_Settings.iAKeyConfig[1]--;
			} else if (sel == KEY_CONF_A_DOWN){
				if (PSP_Settings.iAKeyConfig[2] < 1) {
					PSP_Settings.iAKeyConfig[2] = 1;
				}
				PSP_Settings.iAKeyConfig[2]--;
			} else if (sel == KEY_CONF_SWAPAD){
				PSP_Settings.bSwapAnalog = true;
			} else if (sel == KEY_CONF_SELLOCK) {
				PSP_Settings.bAkeyLock = true;
			} else if (sel == KEY_CONF_WPAD) {
				PSP_Settings.bWPad = true;
			}
		} else if (new_pad & CTRL_RIGHT){
			if (sel == KEY_CONF_A_RIGHT){
				PSP_Settings.iAKeyConfig[0]++;
				if (PSP_Settings.iAKeyConfig[0] > AKEY_TRANS) {
					PSP_Settings.iAKeyConfig[0] = AKEY_TRANS;
				}
			} else if (sel == KEY_CONF_A_UP){
				PSP_Settings.iAKeyConfig[1]++;
				if (PSP_Settings.iAKeyConfig[1] > AKEY_TRANS) {
					PSP_Settings.iAKeyConfig[1] = AKEY_TRANS;
				}
			} else if (sel == KEY_CONF_A_DOWN){
				PSP_Settings.iAKeyConfig[2]++;
				if (PSP_Settings.iAKeyConfig[2] > AKEY_TRANS) {
					PSP_Settings.iAKeyConfig[2] = AKEY_TRANS;
				}
			} else if (sel == KEY_CONF_SWAPAD){
				PSP_Settings.bSwapAnalog = false;
			} else if (sel == KEY_CONF_SELLOCK) {
				PSP_Settings.bAkeyLock = false;
			} else if (sel == KEY_CONF_WPAD) {
				PSP_Settings.bWPad = false;
			}
		}
		
		if(sel>=KEY_CONF_A_RIGHT)
			strcpy(msg,"FI  ~FLZ");
		else
			strcpy(msg," ~   L R START SELECTF{^̐ݒ");
		menu_frame(0, msg);
		
		mh_print(2*8,4*8,(unsigned char*)"  {^ݒ胁j[",RGB(105,105,115));

		x = 2;
		y = 6;
		
		msg[0]=9; msg[1]=0;
		pgPrint(x+4,y,RGB(90,200,245),msg);
		pgPrint(x,y++,0xffff,"       X :");
		pgPrint(x+3,y,RGB(90,245,40),msg);
		pgPrint(x,y++,0xffff,"       Y :");
		pgPrint(x+5,y,RGB(245,40,40),msg);
		pgPrint(x,y++,0xffff,"       A :");
		pgPrint(x+4,y,RGB(245,245,40),msg);
		pgPrint(x,y++,0xffff,"       B :");
		pgPrint(x,y++,0xffff,"       L :");
		pgPrint(x,y++,0xffff,"       R :");
		pgPrint(x,y++,0xffff,"   START :");
		pgPrint(x,y++,0xffff,"  SELECT :");
		y++;
		
		if (PSP_Settings.bSwapAnalog) {
			strcpy(msg,"  fW^     :");
			msg[11]=0x81;
			msg[12]=0xA8;
			mh_print(8*x,8*y++,(unsigned char*)msg,0xffff);
			msg[11]=0x81;
			msg[12]=0xAA;
			mh_print(8*x,8*y++,(unsigned char*)msg,0xffff);
			msg[11]=0x81;
			msg[12]=0xAB;
			mh_print(8*x,8*y++,(unsigned char*)msg,0xffff);
			y++;
			mh_print(8*x,8*y++,(unsigned char*)"  AiOƃfW^̓ւ     : ON",0xffff);
		} else {
			strcpy(msg,"  AiO     :");
			msg[11]=0x81;
			msg[12]=0xA8;
			mh_print(8*x,8*y++,(unsigned char*)msg,0xffff);
			msg[11]=0x81;
			msg[12]=0xAA;
			mh_print(8*x,8*y++,(unsigned char*)msg,0xffff);
			msg[11]=0x81;
			msg[12]=0xAB;
			mh_print(8*x,8*y++,(unsigned char*)msg,0xffff);
			y++;
			mh_print(8*x,8*y++,(unsigned char*)"  AiOƃfW^̓ւ     : OFF",0xffff);
		}
		if (PSP_Settings.bAkeyLock) {
			mh_print(8*x,8*y++,(unsigned char*)"  j[select+LockɊ蓖Ă: ON",0xffff);
			if (PSP_Settings.bWPad)
				mh_print(8*x,8*y++,(unsigned char*)"  AiO/fW^pbh      : ON",0xffff);
			else
				mh_print(8*x,8*y++,(unsigned char*)"  AiO/fW^pbh      : OFF",0xffff);
		} else {
			mh_print(8*x,8*y++,(unsigned char*)"  j[select+LockɊ蓖Ă: OFF",0xffff);
			if (PSP_Settings.bWPad)
				mh_print(8*x,8*y++,(unsigned char*)"  AiO/fW^pbh      : ON",RGB(105,105,115));
			else
				mh_print(8*x,8*y++,(unsigned char*)"  AiO/fW^pbh      : OFF",RGB(105,105,115));
		}
		y++;
		
		mh_print(8*x,8*y++,(unsigned char*)"  {^ݒ̏",0xffff);
		mh_print(8*x,8*y++,(unsigned char*)"  Cj[ɖ߂",0xffff);
		
		for (i=0; i<=KEY_CONF_SELECT; i++) {
			y = i + 6;
			msg[0]=0;
				if (PSP_Settings.iKeyConfig[i] & CTRL_CIRCLE) {
					msg[0]=1; msg[1]=0;
				} else if (PSP_Settings.iKeyConfig[i] & CTRL_CROSS) {
					msg[0]=2; msg[1]=0;
				} else if (PSP_Settings.iKeyConfig[i] & CTRL_TRIANGLE) {
					msg[0]=4; msg[1]=0;
				} else if (PSP_Settings.iKeyConfig[i] & CTRL_SQUARE) {
					msg[0]=3; msg[1]=0;
				} else if (PSP_Settings.iKeyConfig[i] & CTRL_LTRIGGER) {
					msg[0]='L'; msg[1]=0;
				} else if (PSP_Settings.iKeyConfig[i] & CTRL_RTRIGGER) {
					msg[0]='R'; msg[1]=0;
				} else if (PSP_Settings.iKeyConfig[i] & CTRL_START) {
					strcpy(msg,"START");
				} else if (PSP_Settings.iKeyConfig[i] & CTRL_SELECT) {
					strcpy(msg,"SELECT");
				}
			pgPrint(13,y,0xffff,msg);
		}
		
		for (i=0; i<3; i++) {
			y = i + 6 + KEY_CONF_A_RIGHT + 1;
			msg[0]=0;
			switch (PSP_Settings.iAKeyConfig[i]) {
			default:
			case AKEY_NONE:
				strcpy(msg,"蓖ĂȂ");
				break;
			case AKEY_FRAMESKIPUP:
				strcpy(msg,"t[XLbv̑");
				break;
			case AKEY_FRAMESKIPDOWN:
				strcpy(msg,"t[XLbv̌");
				break;
			case AKEY_AUTOSKIP:
				strcpy(msg,"t[XLbv؂ւ");
				break;
			case AKEY_SCRSIZE:
				strcpy(msg,"ʃTCY̕ύX");
				break;
			case AKEY_SHOWFPS:
				strcpy(msg,"FPS\\̐؂ւ");
				break;
			case AKEY_VSYNC:
				strcpy(msg,"VSync");
				break;
			case AKEY_SOUNDOFF:
				strcpy(msg,"TEh̐؂ւ");
				break;
			case AKEY_TRANS:
				strcpy(msg,"߂̐؂ւ");
				break;
			}
			mh_print(8*16,8*y,(unsigned char*)msg,0xffff);
		}
		
		y = sel + 6;
		if (sel >= KEY_CONF_A_RIGHT) y++;
		if (sel >= KEY_CONF_SWAPAD) y++;
		if (sel >= KEY_CONF_INIT) y++;
		pgPutChar((x+1)*8,y*8,0xffff,0,127,1,0,1);
		
		pgScreenFlipV();
	}
}

void load_global_config(void)
{
	memset( &Global_Settings, 0, sizeof( Global_Settings ) );
	
	char CFGPath [_MAX_PATH + 1];
	strcpy(CFGPath,PBPPath);
	strcat(CFGPath,"uo_Snes9x.cfg");
	FILE *fd;
	if ((fd = fopen (CFGPath, "rb"))) {
		fread (&Global_Settings, sizeof(Global_Settings), 1, fd);
		fclose (fd);
		if (Global_Settings.Default.iSaveSlot < 0 || Global_Settings.Default.iSaveSlot > SAVE_SLOT_MAX)
			Global_Settings.Default.iSaveSlot = 0;
		if (Global_Settings.Default.iSkipFrames < 0 || Global_Settings.Default.iSkipFrames > 10)
			Global_Settings.Default.iSkipFrames = 0;
		if (Global_Settings.Default.iSoundRate < 0 || Global_Settings.Default.iSoundRate > 2)
			Global_Settings.Default.iSoundRate = 2;
		if (Global_Settings.Default.iHBlankCycleDiv < 0 || Global_Settings.Default.iHBlankCycleDiv > PSP_CYCLE_DIV_MAX)
			Global_Settings.Default.iHBlankCycleDiv = 10;
		if (Global_Settings.Default.iAPUTimerCycleDiv < 0 || Global_Settings.Default.iAPUTimerCycleDiv > PSP_CYCLE_DIV_MAX)
			Global_Settings.Default.iAPUTimerCycleDiv = 10;
		if (Global_Settings.Default.iPSP_ClockUp < 0 || Global_Settings.Default.iPSP_ClockUp > 2)
			Global_Settings.Default.iPSP_ClockUp = 0;
		if (Global_Settings.Default.iScreenSize < SCR_X1 || Global_Settings.Default.iScreenSize > SCR_MAX)
			Global_Settings.Default.iScreenSize = SCR_X1;
		if (Global_Settings.Default.iKeyConfig[0]==0) {
			Global_Settings.Default.iKeyConfig[0]=CTRL_TRIANGLE;
			Global_Settings.Default.iKeyConfig[1]=CTRL_SQUARE;
			Global_Settings.Default.iKeyConfig[2]=CTRL_CIRCLE;
			Global_Settings.Default.iKeyConfig[3]=CTRL_CROSS;
			Global_Settings.Default.iKeyConfig[4]=CTRL_LTRIGGER;
			Global_Settings.Default.iKeyConfig[5]=CTRL_RTRIGGER;
			Global_Settings.Default.iKeyConfig[6]=CTRL_START;
			Global_Settings.Default.iKeyConfig[7]=CTRL_SELECT;
			Global_Settings.Default.iAKeyConfig[0]=AKEY_AUTOSKIP;
			Global_Settings.Default.iAKeyConfig[1]=AKEY_FRAMESKIPUP;
			Global_Settings.Default.iAKeyConfig[2]=AKEY_FRAMESKIPDOWN;
			Global_Settings.Default.bSwapAnalog = false;
			Global_Settings.Default.bAkeyLock = false;
			Global_Settings.Default.bWPad = false;
		}
		if (Global_Settings.LastPath[0]!='m' && Global_Settings.LastPath[1]!='s' &&
			Global_Settings.LastPath[2]!='0' && Global_Settings.LastPath[3]!=':' &&
			Global_Settings.LastPath[4]!='/' )
			strcpy(Global_Settings.LastPath, PBPPath);
	}
	
	char vercnf[48];
	strcpy(vercnf, PSP_CFG_VERSION);
	if (strcmp(Global_Settings.Default.vercnf, vercnf) != 0) {
		strcpy(Global_Settings.Default.vercnf, vercnf);
		Global_Settings.Default.iSaveSlot = 0;
		Global_Settings.Default.iSkipFrames = 0;
		Global_Settings.Default.bShowFPS = false;
		Global_Settings.Default.bVSync = false;
		Global_Settings.Default.bSoundOff = false;
		Global_Settings.Default.iSoundRate = 2;
		Global_Settings.Default.bTrans = true;
		Global_Settings.Default.iHBlankCycleDiv = 10;
		Global_Settings.Default.iAPUTimerCycleDiv = 10;
		Global_Settings.Default.bSaveThumb = true;
		Global_Settings.Default.iPSP_ClockUp = 0;
		Global_Settings.Default.iScreenSize = SCR_X1;
		Global_Settings.Default.bAutoSkip = true;
		Global_Settings.Default.iKeyConfig[0]=CTRL_TRIANGLE;
		Global_Settings.Default.iKeyConfig[1]=CTRL_SQUARE;
		Global_Settings.Default.iKeyConfig[2]=CTRL_CIRCLE;
		Global_Settings.Default.iKeyConfig[3]=CTRL_CROSS;
		Global_Settings.Default.iKeyConfig[4]=CTRL_LTRIGGER;
		Global_Settings.Default.iKeyConfig[5]=CTRL_RTRIGGER;
		Global_Settings.Default.iKeyConfig[6]=CTRL_START;
		Global_Settings.Default.iKeyConfig[7]=CTRL_SELECT;
		Global_Settings.Default.iAKeyConfig[0]=AKEY_AUTOSKIP;
		Global_Settings.Default.iAKeyConfig[1]=AKEY_FRAMESKIPUP;
		Global_Settings.Default.iAKeyConfig[2]=AKEY_FRAMESKIPDOWN;
		Global_Settings.Default.bSwapAnalog = false;
		Global_Settings.Default.bAkeyLock = false;
		Global_Settings.Default.bWPad = false;
		strcpy(Global_Settings.LastPath, PBPPath);
	}
	strcpy(LastPath, Global_Settings.LastPath);
}

void save_config(void)
{
	char CFGPath [_MAX_PATH + 1];
	strcpy(CFGPath, S9xGetFilename("cfg"));
	FILE *fd;
	if ((fd = fopen (CFGPath, "wb"))) {
		fwrite (&PSP_Settings, sizeof(PSP_Settings), 1, fd);
		fclose (fd);
	}
	memset(Global_Settings.LastPath, 0, MAX_PATH);
	strcpy(Global_Settings.LastPath, LastPath);
	strcpy(CFGPath,PBPPath);
	strcat(CFGPath,"uo_Snes9x.cfg");
	if ((fd = fopen (CFGPath, "wb"))) {
		fwrite (&Global_Settings, sizeof(Global_Settings), 1, fd);
		fclose (fd);
	}
}

void load_config(void)
{
	if (!PSP_Settings.bSoundOff) {
		S9xSetSoundMute( TRUE );
		if ( g_thread !=-1 ){
			Settings.ThreadSound = FALSE;
			sceKernelWaitThreadEnd( g_thread, NULL );
			sceKernelDeleteThread( g_thread );
			g_thread = -1;
		}
	}
	
	memset( &PSP_Settings, 0, sizeof( PSP_Settings ) );
	
	char CFGPath [_MAX_PATH + 1];
	strcpy(CFGPath, S9xGetFilename("cfg"));
	FILE *fd;
	if ((fd = fopen (CFGPath, "rb"))) {
		fread (&PSP_Settings, sizeof(PSP_Settings), 1, fd);
		fclose (fd);
		
		if (PSP_Settings.iSaveSlot < 0 || PSP_Settings.iSaveSlot > SAVE_SLOT_MAX)
			PSP_Settings.iSaveSlot = 0;
		if (PSP_Settings.iSkipFrames < 0 || PSP_Settings.iSkipFrames > 10)
			PSP_Settings.iSkipFrames = 0;
		if (PSP_Settings.iSoundRate < 0 || PSP_Settings.iSoundRate > 2)
			PSP_Settings.iSoundRate = 2;
		if (PSP_Settings.iHBlankCycleDiv < 0 || PSP_Settings.iHBlankCycleDiv > PSP_CYCLE_DIV_MAX)
			PSP_Settings.iHBlankCycleDiv = 10;
		if (PSP_Settings.iAPUTimerCycleDiv < 0 || PSP_Settings.iAPUTimerCycleDiv > PSP_CYCLE_DIV_MAX)
			PSP_Settings.iAPUTimerCycleDiv = 10;
		if (PSP_Settings.iPSP_ClockUp < 0 || PSP_Settings.iPSP_ClockUp > 2)
			PSP_Settings.iPSP_ClockUp = 0;
		if (PSP_Settings.iScreenSize < SCR_X1 || PSP_Settings.iScreenSize > SCR_MAX)
			PSP_Settings.iScreenSize = SCR_X1;
		if (PSP_Settings.iKeyConfig[0]==0)
			key_config_init();
	}
	
	char vercnf[48];
	strcpy(vercnf, PSP_CFG_VERSION);
	if (strcmp(PSP_Settings.vercnf, vercnf) != 0) {
		PSP_Settings = Global_Settings.Default;
		strcpy(PSP_Settings.vercnf, vercnf);
	}
	
	if (PSP_Settings.iSoundRate==0) S9xSetPlaybackRate( 11025 );
	else if (PSP_Settings.iSoundRate==1) S9xSetPlaybackRate( 22050 );
	else S9xSetPlaybackRate( 44100 );
	if (!PSP_Settings.bSoundOff) {
		Settings.ThreadSound = TRUE;
		InitTimer();
	}
	Settings.H_Max = (long)((SNES_CYCLES_PER_SCANLINE * 1000) / ( PSP_Settings.iHBlankCycleDiv * 100));
	Settings.HBlankStart = ((uint32)Settings.H_Max << 8) / SNES_HCOUNTER_MAX;
	fix_H_Max();
	Settings.PSP_APUTimerCycle = (uint32)((SNES_APUTIMER2_CYCLEx10000 * 1000) / (PSP_Settings.iAPUTimerCycleDiv * 100));
	if (PSP_Settings.bTrans){
		Settings.Transparency = TRUE;
	} else {
		Settings.Transparency = FALSE;
	}
	Settings.SkipFrames = PSP_Settings.iSkipFrames;
#ifdef RELEASE
	if (PSP_Settings.iScreenSize == SCR_X1) {
		GFX.Screen = (uint8*)pgGetVramAddr( (SCREEN_WIDTH - SNES_WIDTH) >> 1, (SCREEN_HEIGHT - SNES_HEIGHT) >> 1 );
	} else {
		GFX.Screen = (uint8*)GFX_Screen0;
		GeFlip = 0;
	}
	if (PSP_Settings.iScreenSize >= SCR_GEFIT) {
		Settings.SupportHiRes = TRUE;
	} else {
		Settings.SupportHiRes = FALSE;
	}
#endif // RELEASE
	if (PSP_Settings.iPSP_ClockUp == 2) scePowerSetClockFrequency(333,333,166);
	else if (PSP_Settings.iPSP_ClockUp == 1) scePowerSetClockFrequency(266,266,133);
	else scePowerSetClockFrequency(222,222,111);
}

void open_menu(void)
{
	enum {
		SRAM_SAVE = 0,
		STATE_SLOT,
		STATE_SAVE,
		STATE_LOAD,
		STATE_DEL,
		SAVE_THUMB,

		FRAME_SKIP,
		AUTO_SKIP,
		SCR_SIZE,
		SHOWFPS,
		VSYNC,
		SOUNDOFF,
		SOUNDRATE,
		TRANS,
		HBLANK_CYCLE,
		APU_CYCLE,
		PSP_CLOCKUP,
		KEY_CONFIG,

		LOAD_ROM,
		RESET,
		EXIT_HOME,
		CONTINUE,
	};
	char msg[256], tmp[256];
	static int sel=0;
	int x, y;
	int ret;
	
	get_screenshot((unsigned char *)GFX.Screen);
	scePowerSetClockFrequency(222,222,111);
	get_slotdate(S9xGetFilename("sv0"));
	get_thumbs(S9xGetFilename("tn0"));
	
	bool f_bDialog = false;
	bool f_bExit = false;
	bool f_bOldSoundOff = PSP_Settings.bSoundOff;
	int f_iOldSoundRate = PSP_Settings.iSoundRate;
	int tmp_color;
	
	old_pad = CTRL_LEFT;
	
	msg[0]=0;
	for(;;){
		readpad();
		
		if (new_pad & CTRL_CIRCLE){
			if (sel == SRAM_SAVE){
				if ( Memory.SaveSRAM( S9xGetFilename("srm") ) ) {
						strcpy(msg, "SRAMۑɐ܂");
				} else {
						strcpy(msg, "SRAMۑɎsSRAM܂");
				}
			} else if (sel == STATE_SLOT){
				PSP_Settings.iSaveSlot++;
				if (PSP_Settings.iSaveSlot > SAVE_SLOT_MAX) {
					PSP_Settings.iSaveSlot = 0;
				}
			} else if (sel == STATE_SAVE){
state_save: ;
				pgFillBox( 129, 104, 351, 168, 0 );
				strcpy(tmp,"{ɃXbg 0 ɕۑ܂H");
				tmp[strlen(tmp)-18] = PSP_Settings.iSaveSlot + '0';
				mh_print(165, 122, (unsigned char*)tmp, RGB(90,200,245));
				mh_print(195, 142, (unsigned char*)"FOK  ~FLZ", RGB(105,105,115));
				pgScreenFlipV();
				for(;;){
					readpad();
					if (new_pad & CTRL_CIRCLE){
						f_bDialog = true;
						break;
					} else if (new_pad & CTRL_CROSS){
						f_bDialog = false;
						break;
					}
					pgWaitV();
				}
				if (f_bDialog) {
					pgFillBox( 129, 104, 351, 168, 0 );
					mh_print(195, 132, (unsigned char*)"ԕۑĂ܂...", RGB(255,205,0));
					pgScreenFlipV();
					if ( S9xFreezeGame( S9xGetFilename("sv0") ) ) {
							if (PSP_Settings.bSaveThumb) {
								save_thumb(S9xGetFilename("tn0"));
							} else {
								delete_file(S9xGetFilename("tn0"));
							}
							strcpy(msg, "ԕۑɐ܂B");
							get_slotdate(S9xGetFilename("sv0"));
							get_thumbs(S9xGetFilename("tn0"));
					} else {
							strcpy(msg, "ԕۑɎs܂B");
					}
					S9xSetSoundMute( TRUE );
				}
			} else if (sel == STATE_LOAD){
				if (slotflag[PSP_Settings.iSaveSlot]) {
					pgFillBox( 129, 104, 351, 168, 0 );
					strcpy(tmp,"{ɃXbg 0 ǂݍ݂܂H");
					tmp[strlen(tmp)-22] = PSP_Settings.iSaveSlot + '0';
					mh_print(165, 122, (unsigned char*)tmp, RGB(90,245,40));
					mh_print(195, 142, (unsigned char*)"FOK  ~FLZ", RGB(105,105,115));
					pgScreenFlipV();
					for(;;){
						readpad();
						if (new_pad & CTRL_CIRCLE){
							f_bDialog = true;
							break;
						} else if (new_pad & CTRL_CROSS){
							f_bDialog = false;
							break;
						}
						pgWaitV();
					}
					if (f_bDialog) {
						if ( S9xUnfreezeGame( S9xGetFilename("sv0") ) ) {
							S9xSetInfoString( "ԓǂݍ݂ɐ܂" );
							break;
						} else {
							strcpy(msg, "ԕۑɎs܂");
						}
					}
				}
			} else if (sel == STATE_DEL){
				if (slotflag[PSP_Settings.iSaveSlot]) {
					pgFillBox( 129, 104, 351, 168, 0 );
					strcpy(tmp,"{ɃXbg 0 폜܂H");
					tmp[strlen(tmp)-18] = PSP_Settings.iSaveSlot + '0';
					mh_print(160, 122, (unsigned char*)tmp, RGB(245,245,40));
					mh_print(195, 142, (unsigned char*)"FOK  ~FLZ", RGB(105,105,115));
					pgScreenFlipV();
					for(;;){
						readpad();
						if (new_pad & CTRL_CIRCLE){
							f_bDialog = true;
							break;
						} else if (new_pad & CTRL_CROSS){
							f_bDialog = false;
							break;
						}
						pgWaitV();
					}
					if (f_bDialog) {
						delete_file(S9xGetFilename("sv0"));
						delete_file(S9xGetFilename("tn0"));
						get_slotdate(S9xGetFilename("sv0"));
						get_thumbs(S9xGetFilename("tn0"));
					}
				}
			} else if (sel == SAVE_THUMB){
				PSP_Settings.bSaveThumb = !PSP_Settings.bSaveThumb;
			} else if (sel == FRAME_SKIP){
				PSP_Settings.iSkipFrames++;
				if ( PSP_Settings.iSkipFrames > 10 ){
					PSP_Settings.iSkipFrames = 0;
				}
			} else if (sel == AUTO_SKIP){
				PSP_Settings.bAutoSkip = !PSP_Settings.bAutoSkip;
			} else if (sel == SCR_SIZE){
				PSP_Settings.iScreenSize++;
				if (PSP_Settings.iScreenSize > SCR_MAX) {
					PSP_Settings.iScreenSize = SCR_X1;
				}
			} else if (sel == SHOWFPS){
				PSP_Settings.bShowFPS = !PSP_Settings.bShowFPS;
			} else if (sel == VSYNC){
				PSP_Settings.bVSync = !PSP_Settings.bVSync;
			} else if (sel == SOUNDOFF){
				PSP_Settings.bSoundOff = !PSP_Settings.bSoundOff;
			} else if (sel == SOUNDRATE){
				PSP_Settings.iSoundRate++;
				if (PSP_Settings.iSoundRate > 2) {
					PSP_Settings.iSoundRate = 0;
				}
			} else if (sel == TRANS){
				PSP_Settings.bTrans = !PSP_Settings.bTrans;
			} else if (sel == HBLANK_CYCLE){
				PSP_Settings.iHBlankCycleDiv++;
				if (PSP_Settings.iHBlankCycleDiv > PSP_CYCLE_DIV_MAX) {
					PSP_Settings.iHBlankCycleDiv = 10;
				}
			} else if (sel == APU_CYCLE){
				PSP_Settings.iAPUTimerCycleDiv++;
				if (PSP_Settings.iAPUTimerCycleDiv > PSP_CYCLE_DIV_MAX) {
					PSP_Settings.iAPUTimerCycleDiv = 10;
				}
			} else if (sel == KEY_CONFIG){
				key_config();
			} else if (sel == PSP_CLOCKUP){
				PSP_Settings.iPSP_ClockUp++;
				if (PSP_Settings.iPSP_ClockUp > 2) {
					PSP_Settings.iPSP_ClockUp = 0;
				}
			} else if (sel == LOAD_ROM){
				msg[0]=0;
				FilerMsg[0]=0;
				if (getFilePath(RomPath)){
					pgFillBox( 129, 104, 351, 168, 0 );
					mh_print(205, 132, (unsigned char*)"ǂݍݒł...", RGB(90,200,245));
					pgScreenFlipV();
					save_config();
					Memory.SaveSRAM( S9xGetFilename("srm") );
					if ( Memory.LoadROM(RomPath) ){
						Memory.LoadSRAM( S9xGetFilename("srm") );
						load_config();
						S9xSetInfoString( "ROMC[Wǂݍ݂܂" );
						return;
					} else {
						strcpy(msg, "RomC[W̓ǂݍ݂Ɏs܂");
					}
				}
			} else if (sel == RESET){
				S9xReset();
				S9xSetInfoString( "Q[Zbg܂" );
				break;
			} else if (sel == EXIT_HOME){
				f_bExit = true;
				break;
			} else if (sel == CONTINUE){
				break;
			}
		} else if (new_pad & CTRL_CROSS){
			break;
		} else if (new_pad & CTRL_TRIANGLE){
			goto state_save;
		} else if (new_pad & CTRL_SQUARE){
			Global_Settings.Default = PSP_Settings;
			strcpy(msg, "݂̐ݒftHgɂ܂");
		} else if (new_pad & CTRL_UP){
			if (sel!=0){
				sel--;
			} else {
				sel=CONTINUE;
			}
		} else if (new_pad & CTRL_DOWN){
			if (sel!=CONTINUE){
				sel++;
			} else {
				sel=0;
			}
		} else if (new_pad & CTRL_LEFT){
			if (sel == STATE_SLOT || sel == STATE_SAVE || sel == STATE_LOAD || sel == STATE_DEL){
				if (PSP_Settings.iSaveSlot < 1) {
					PSP_Settings.iSaveSlot = SAVE_SLOT_MAX + 1;
				}
				PSP_Settings.iSaveSlot--;
			} else if (sel == SAVE_THUMB){
				PSP_Settings.bSaveThumb = !PSP_Settings.bSaveThumb;
			} else if (sel == FRAME_SKIP){
				if ( PSP_Settings.iSkipFrames > 0 ){
					PSP_Settings.iSkipFrames--;
				}
			} else if (sel == AUTO_SKIP){
				PSP_Settings.bAutoSkip = true;
			} else if (sel == SCR_SIZE){
				if (PSP_Settings.iScreenSize < 1) {
					PSP_Settings.iScreenSize = 1;
				}
				PSP_Settings.iScreenSize--;
			} else if (sel == SHOWFPS){
				PSP_Settings.bShowFPS = true;
			} else if (sel == VSYNC){
				PSP_Settings.bVSync = true;
			} else if (sel == SOUNDOFF){
				PSP_Settings.bSoundOff = false;
			} else if (sel == SOUNDRATE){
				if (PSP_Settings.iSoundRate < 1) {
					PSP_Settings.iSoundRate = 1;
				}
				PSP_Settings.iSoundRate--;
			} else if (sel == TRANS){
				PSP_Settings.bTrans = true;
			} else if (sel == HBLANK_CYCLE){
				if (PSP_Settings.iHBlankCycleDiv < 11) {
					PSP_Settings.iHBlankCycleDiv = 11;
				}
				PSP_Settings.iHBlankCycleDiv--;
			} else if (sel == APU_CYCLE){
				if (PSP_Settings.iAPUTimerCycleDiv < 11) {
					PSP_Settings.iAPUTimerCycleDiv = 11;
				}
				PSP_Settings.iAPUTimerCycleDiv--;
			} else if (sel == PSP_CLOCKUP){
				if (PSP_Settings.iPSP_ClockUp < 1) {
					PSP_Settings.iPSP_ClockUp = 1;
				}
				PSP_Settings.iPSP_ClockUp--;
			}
		} else if (new_pad & CTRL_RIGHT){
			if (sel == STATE_SLOT || sel == STATE_SAVE || sel == STATE_LOAD || sel == STATE_DEL){
				PSP_Settings.iSaveSlot++;
				if (PSP_Settings.iSaveSlot > SAVE_SLOT_MAX) {
					PSP_Settings.iSaveSlot = 0;
				}
			} else if (sel == SAVE_THUMB){
				PSP_Settings.bSaveThumb = !PSP_Settings.bSaveThumb;
			} else if (sel == FRAME_SKIP){
				if ( PSP_Settings.iSkipFrames < 10 ){
					PSP_Settings.iSkipFrames++;
				}
			} else if (sel == AUTO_SKIP){
				PSP_Settings.bAutoSkip = false;
			} else if (sel == SCR_SIZE){
				PSP_Settings.iScreenSize++;
				if (PSP_Settings.iScreenSize > SCR_MAX) {
					PSP_Settings.iScreenSize = SCR_MAX;
				}
			} else if (sel == SHOWFPS){
				PSP_Settings.bShowFPS = false;
			} else if (sel == VSYNC){
				PSP_Settings.bVSync = false;
			} else if (sel == SOUNDOFF){
				PSP_Settings.bSoundOff = true;
			} else if (sel == SOUNDRATE){
				PSP_Settings.iSoundRate++;
				if (PSP_Settings.iSoundRate > 2) {
					PSP_Settings.iSoundRate = 2;
				}
			} else if (sel == TRANS){
				PSP_Settings.bTrans = false;
			} else if (sel == HBLANK_CYCLE){
				PSP_Settings.iHBlankCycleDiv++;
				if (PSP_Settings.iHBlankCycleDiv > PSP_CYCLE_DIV_MAX) {
					PSP_Settings.iHBlankCycleDiv = PSP_CYCLE_DIV_MAX;
				}
			} else if (sel == APU_CYCLE){
				PSP_Settings.iAPUTimerCycleDiv++;
				if (PSP_Settings.iAPUTimerCycleDiv > PSP_CYCLE_DIV_MAX) {
					PSP_Settings.iAPUTimerCycleDiv = PSP_CYCLE_DIV_MAX;
				}
			} else if (sel == PSP_CLOCKUP){
				PSP_Settings.iPSP_ClockUp++;
				if (PSP_Settings.iPSP_ClockUp > 2) {
					PSP_Settings.iPSP_ClockUp = 2;
				}
			}
		} else if ((now_pad & CTRL_SELECT) && (now_pad & CTRL_START)) {
			f_bExit = true;
			break;
		} else if (new_pad & CTRL_LTRIGGER){
			if (sel>LOAD_ROM) {
				sel=LOAD_ROM;
			} else if (sel>FRAME_SKIP){
				sel=FRAME_SKIP;
			} else if (sel>SRAM_SAVE){
				sel=SRAM_SAVE;
			}
		} else if (new_pad & CTRL_RTRIGGER){
			if (sel<FRAME_SKIP) {
				sel=FRAME_SKIP;
			} else if (sel<LOAD_ROM){
				sel=LOAD_ROM;
			} else if (sel<CONTINUE){
				sel=CONTINUE;
			}
		}
		
		menu_frame(msg, ":OK ~:LZ :ۑ :lɐݒ SELECT+START:I");
		
		mh_print(33, 33, (unsigned char*)Memory.ROMFilename, RGB(110,115,175));
		
		x = 2;
		y = 6;
		
		mh_print(8*x,8*y++,(unsigned char*)"  SRAMۑ",0xffff);
		strcpy(tmp,"  ۑXbg        : 0");
		tmp[strlen(tmp)-1] = PSP_Settings.iSaveSlot + '0';
		mh_print(8*x,8*y++,(unsigned char*)tmp,0xffff);
		
		mh_print(8*x,8*y++,(unsigned char*)"  ԕۑ",0xffff);
		mh_print(8*x,8*y++,(unsigned char*)"  ԓǂݍ",0xffff);
		mh_print(8*x,8*y++,(unsigned char*)"  ԍ폜",0xffff);
		if (PSP_Settings.bSaveThumb) {
			mh_print(8*x,8*y++,(unsigned char*)"  k摜̕ۑ      : ON",0xffff);
		} else {
			mh_print(8*x,8*y++,(unsigned char*)"  k摜̕ۑ      : OFF",0xffff);
		}
		y++;
		
		strcpy(tmp,"  t[XLbv    : 00");
		tmp[strlen(tmp)-2] = PSP_Settings.iSkipFrames / 10 + '0';
		tmp[strlen(tmp)-1] = PSP_Settings.iSkipFrames % 10 + '0';
		mh_print(8*x,8*y++,(unsigned char*)tmp,0xffff);
		
		if (PSP_Settings.bAutoSkip) {
			mh_print(8*x,8*y++,(unsigned char*)"  t[XLbv: ON",0xffff);
		} else {
			mh_print(8*x,8*y++,(unsigned char*)"  t[XLbv: OFF",0xffff);
		}
		if (PSP_Settings.iScreenSize==SCR_X1) {
			mh_print(8*x,8*y++,(unsigned char*)"  ʃTCY          : W",0xffff);
		} else if (PSP_Settings.iScreenSize==SCR_FIT){
			mh_print(8*x,8*y++,(unsigned char*)"  ʃTCY          : tBbg",0xffff);
		} else if (PSP_Settings.iScreenSize==SCR_43FIT){
			mh_print(8*x,8*y++,(unsigned char*)"  ʃTCY          : 4:3",0xffff);
		} else if (PSP_Settings.iScreenSize==SCR_WX2){
			mh_print(8*x,8*y++,(unsigned char*)"  ʃTCY          : {",0xffff);
		} else if (PSP_Settings.iScreenSize==SCR_FULLFIT){
			mh_print(8*x,8*y++,(unsigned char*)"  ʃTCY          : S",0xffff);
		} else if (PSP_Settings.iScreenSize==SCR_GEFIT){
			mh_print(8*x,8*y++,(unsigned char*)"  ʃTCY          : tBbg(BL)",0xffff);
		} else if (PSP_Settings.iScreenSize==SCR_GE43FIT){
			mh_print(8*x,8*y++,(unsigned char*)"  ʃTCY          : 4:3 (BL)",0xffff);
		} else if (PSP_Settings.iScreenSize==SCR_GEFULLFIT){
			mh_print(8*x,8*y++,(unsigned char*)"  ʃTCY          : S(BL)",0xffff);
		}
		if (PSP_Settings.bShowFPS) {
			mh_print(8*x,8*y++,(unsigned char*)"  FPS̕\\           : ON",0xffff);
		} else {
			mh_print(8*x,8*y++,(unsigned char*)"  FPS̕\\           : OFF",0xffff);
		}
		if (PSP_Settings.bVSync) {
			mh_print(8*x,8*y++,(unsigned char*)"  VSync               : ON",0xffff);
		} else {
			mh_print(8*x,8*y++,(unsigned char*)"  VSync               : OFF",0xffff);
		}
		if (PSP_Settings.bSoundOff) {
			mh_print(8*x,8*y++,(unsigned char*)"  TEh            : OFF",0xffff);
		} else {
			mh_print(8*x,8*y++,(unsigned char*)"  TEh            : ON",0xffff);
		}
		if (PSP_Settings.iSoundRate==2) {
			mh_print(8*x,8*y++,(unsigned char*)"  TEh[g      : 44kHz",0xffff);
		} else if (PSP_Settings.iSoundRate==1){
			mh_print(8*x,8*y++,(unsigned char*)"  TEh[g      : 22kHz",0xffff);
		} else {
			mh_print(8*x,8*y++,(unsigned char*)"  TEh[g      : 11kHz",0xffff);
		}
		if (PSP_Settings.bTrans) {
			mh_print(8*x,8*y++,(unsigned char*)"                  : ON",0xffff);
		} else {
			mh_print(8*x,8*y++,(unsigned char*)"                  : OFF",0xffff);
		}
		strcpy(tmp,"x0.0");
		tmp[strlen(tmp)-3] = PSP_Settings.iHBlankCycleDiv / 10 + '0';
		tmp[strlen(tmp)-1] = PSP_Settings.iHBlankCycleDiv % 10 + '0';
		mh_print(x*8,y*8,(unsigned char*)"  ʑx            : ",0xffff);
		tmp_color = 255-((PSP_Settings.iHBlankCycleDiv - 10) * 10);
		pgPrint(x+15,y++,RGB(255,tmp_color,tmp_color),tmp);
		strcpy(tmp,"x0.0");
		tmp[strlen(tmp)-3] = PSP_Settings.iAPUTimerCycleDiv / 10 + '0';
		tmp[strlen(tmp)-1] = PSP_Settings.iAPUTimerCycleDiv % 10 + '0';
		mh_print(8*x,8*y,(unsigned char*)"  TEhx        : ",0xffff);
		tmp_color = 255-((PSP_Settings.iAPUTimerCycleDiv - 10) * 10);
		pgPrint(x+15,y++,RGB(255,tmp_color,tmp_color),tmp);
		mh_print(8*x,8*y,(unsigned char*)"  PSP̃NbN       : ",0xffff);
		if (PSP_Settings.iPSP_ClockUp == 2) pgPrint(x+15,y++,RGB(255,95,95),"333MHz");
		else if (PSP_Settings.iPSP_ClockUp == 1) pgPrint(x+15,y++,RGB(255,255,95),"266MHz");
		else pgPrint(x+15,y++,0xffff,"222MHz");
		mh_print(8*x,8*y++,(unsigned char*)"  {^ݒ",0xffff);
		y++;
		
		mh_print(8*x,8*y++,(unsigned char*)"  ROMIɖ߂",0xffff);
		mh_print(8*x,8*y++,(unsigned char*)"  ROM̃Zbg",0xffff);
		mh_print(8*x,8*y++,(unsigned char*)"  G~̏I",0xffff);
		mh_print(8*x,8*y++,(unsigned char*)"  Q[̍ĊJ",0xffff);
		
		y = sel + 6;
		if (sel >= FRAME_SKIP) y++;
		if (sel >= LOAD_ROM) y++;
		pgPutChar((x+1)*8,y*8,0xffff,0,127,1,0,1);
		
		if (slotflag[PSP_Settings.iSaveSlot] && thumbflag[PSP_Settings.iSaveSlot]) {
			pgDrawFrame(327,129,456,242,RGB(85,85,95));
			pgBitBlt(328,130,128,112,1,slot_thumb[PSP_Settings.iSaveSlot]);
		}
		mh_print(8*27,8*6,(unsigned char*)"ԕۑꗗ",RGB(105,105,115));
		for(y=0; y<=SAVE_SLOT_MAX; y++){
			if (y==PSP_Settings.iSaveSlot) {
				pgPrint(28,y+7,0xffff,slotdate[y]);
			} else {
				pgPrint(28,y+7,RGB(105,105,115),slotdate[y]);
			}
		}
		
		pgScreenFlipV();
	}
	
	if (f_bExit) {
		pgFillBox( 129, 104, 351, 168, 0 );
		mh_print(210, 132, (unsigned char*)"Ił...", 0xffff);
		pgScreenFlipV();
		save_config();
		Memory.SaveSRAM( S9xGetFilename("srm") );
		if ( g_thread !=-1 ){
			Settings.ThreadSound = FALSE;
			sceKernelWaitThreadEnd( g_thread, NULL );
			sceKernelDeleteThread( g_thread );
		}
		S9xCloseSoundDevice();
		// Exit game
		sceKernelExitGame();
	}
	
	if ((PSP_Settings.iSoundRate != f_iOldSoundRate) || ( PSP_Settings.bSoundOff != f_bOldSoundOff)) {
		if (!f_bOldSoundOff) {
			S9xSetSoundMute( TRUE );
			if ( g_thread !=-1 ){
				Settings.ThreadSound = FALSE;
				sceKernelWaitThreadEnd( g_thread, NULL );
				sceKernelDeleteThread( g_thread );
				g_thread = -1;
			}
		}
		if (PSP_Settings.iSoundRate==0) S9xSetPlaybackRate( 11025 );
		else if (PSP_Settings.iSoundRate==1) S9xSetPlaybackRate( 22050 );
		else S9xSetPlaybackRate( 44100 );
		if (!PSP_Settings.bSoundOff) {
			Settings.ThreadSound = TRUE;
			InitTimer();
		}
	}
	
	Settings.H_Max = (long)((SNES_CYCLES_PER_SCANLINE * 1000) / ( PSP_Settings.iHBlankCycleDiv * 100));
	Settings.HBlankStart = ((uint32)Settings.H_Max << 8) / SNES_HCOUNTER_MAX;
	fix_H_Max();
	Settings.PSP_APUTimerCycle = (uint32)((SNES_APUTIMER2_CYCLEx10000 * 1000) / (PSP_Settings.iAPUTimerCycleDiv * 100));
	
	if (PSP_Settings.bTrans){
		Settings.Transparency = TRUE;
	} else {
		Settings.Transparency = FALSE;
	}
	Settings.SkipFrames = PSP_Settings.iSkipFrames;
#ifdef RELEASE
	if (PSP_Settings.iScreenSize == SCR_X1) {
		GFX.Screen = (uint8*)pgGetVramAddr( (SCREEN_WIDTH - SNES_WIDTH) >> 1, (SCREEN_HEIGHT - SNES_HEIGHT) >> 1 );
	} else {
		GFX.Screen = (uint8*)GFX_Screen0;
		GeFlip = 0;
	}
	if (PSP_Settings.iScreenSize >= SCR_GEFIT) {
		Settings.SupportHiRes = TRUE;
	} else {
		Settings.SupportHiRes = FALSE;
	}
#endif // RELEASE
	if (PSP_Settings.iPSP_ClockUp == 2) scePowerSetClockFrequency(333,333,166);
	else if (PSP_Settings.iPSP_ClockUp == 1) scePowerSetClockFrequency(266,266,133);
	else scePowerSetClockFrequency(222,222,111);
}

void S9xProcessEvents( bool8 block )
{
	if ( s_iAnalog == ANALOG_NONE ) {
		s_iOld_Analog = ANALOG_NONE;
		return;
	}
	
	if ( s_iAnalog == ANALOG_LEFT ){
		S9xSetSoundMute( TRUE );
	
		open_menu();
	
		pgFillvram(0);
		pgScreenFlipV();
		pgFillvram(0);
		pgScreenFlipV();
		pgWaitVn(8);
		if (!PSP_Settings.bSoundOff) { S9xSetSoundMute( FALSE );}
		next1.tv_sec = 0;
	} else {
		if (s_iOld_Analog == s_iAnalog) {
			return;
		}
		s_iOld_Analog = s_iAnalog;
		switch (PSP_Settings.iAKeyConfig[s_iAnalog-2]) {
		default:
		case AKEY_NONE:
			break;
		case AKEY_FRAMESKIPUP:
			if ( PSP_Settings.iSkipFrames < 10 ){
				PSP_Settings.iSkipFrames++;
			}
			if ( Settings.SkipFrames != PSP_Settings.iSkipFrames ){
				Settings.SkipFrames = PSP_Settings.iSkipFrames;
				strcpy( String, "t[XLbv:" );
				if ( Settings.SkipFrames < 10 ){
					String[10] = (Settings.SkipFrames % 10) + '0';
					String[11] = 0;
				} else {
					String[10] = (Settings.SkipFrames / 10) + '0';
					String[11] = (Settings.SkipFrames % 10) + '0';
					String[12] = 0;
				}
				S9xSetInfoString( String );
			}
			break;
		case AKEY_FRAMESKIPDOWN:
			if ( PSP_Settings.iSkipFrames > 0 ){
				PSP_Settings.iSkipFrames--;
			}
			if ( Settings.SkipFrames != PSP_Settings.iSkipFrames ){
				Settings.SkipFrames = PSP_Settings.iSkipFrames;
				strcpy( String, "t[XLbv:" );
				if ( Settings.SkipFrames < 10 ){
					String[10] = (Settings.SkipFrames % 10) + '0';
					String[11] = 0;
				} else {
					String[10] = (Settings.SkipFrames / 10) + '0';
					String[11] = (Settings.SkipFrames % 10) + '0';
					String[12] = 0;
				}
				S9xSetInfoString( String );
			}
			break;
		case AKEY_AUTOSKIP:
			PSP_Settings.bAutoSkip = !PSP_Settings.bAutoSkip;
			if (PSP_Settings.bAutoSkip)
				S9xSetInfoString( "t[XLbv: ON" );
			else
				S9xSetInfoString( "t[XLbv: OFF" );
			break;
		case AKEY_SCRSIZE:
			PSP_Settings.iScreenSize++;
			if (PSP_Settings.iScreenSize > SCR_MAX) {
				PSP_Settings.iScreenSize = SCR_X1;
			}
#ifdef RELEASE
			if (PSP_Settings.iScreenSize == SCR_X1) {
				GFX.Screen = (uint8*)pgGetVramAddr( (SCREEN_WIDTH - SNES_WIDTH) >> 1, (SCREEN_HEIGHT - SNES_HEIGHT) >> 1 );
			} else {
				GFX.Screen = (uint8*)GFX_Screen0;
				GeFlip = 0;
			}
			if (PSP_Settings.iScreenSize >= SCR_GEFIT) {
				Settings.SupportHiRes = TRUE;
			} else {
				Settings.SupportHiRes = FALSE;
			}
			pgFillvram(0);
			pgScreenFlipV();
			pgFillvram(0);
			pgScreenFlipV();
#endif // RELEASE
			break;
		case AKEY_SHOWFPS:
			PSP_Settings.bShowFPS = !PSP_Settings.bShowFPS;
			pgFillvram(0);
			pgScreenFlipV();
			pgFillvram(0);
			pgScreenFlipV();
			break;
		case AKEY_VSYNC:
			PSP_Settings.bVSync = !PSP_Settings.bVSync;
			if (PSP_Settings.bVSync)
				S9xSetInfoString( "VSync: ON" );
			else
				S9xSetInfoString( "VSync: OFF" );
			break;
		case AKEY_SOUNDOFF:
			PSP_Settings.bSoundOff = !PSP_Settings.bSoundOff;
			if (PSP_Settings.bSoundOff) {
				S9xSetSoundMute( TRUE );
				if ( g_thread !=-1 ){
					Settings.ThreadSound = FALSE;
					sceKernelWaitThreadEnd( g_thread, NULL );
					sceKernelDeleteThread( g_thread );
				}
				S9xSetInfoString( "TEh: OFF" );
			} else {
				Settings.ThreadSound = TRUE;
				InitTimer();
				S9xSetSoundMute( FALSE );
				S9xSetInfoString( "TEh: ON" );
			}
			break;
		case AKEY_TRANS:
			PSP_Settings.bTrans = !PSP_Settings.bTrans;
			if (PSP_Settings.bTrans) {
				Settings.Transparency = TRUE;
				S9xSetInfoString( ": ON" );
			} else {
				Settings.Transparency = FALSE;
				S9xSetInfoString( ": OFF" );
			}
			break;
		}
	}
	s_iAnalog = ANALOG_NONE;
}

int xmain(int argc, char *argv)
{
	char *p, savedir[MAX_PATH];
	strcpy(PBPPath, argv);
	p = strrchr(PBPPath, '/');
	*++p = 0;

	pgInit();
	pgScreenFrame(2,0);
	SetupCallbacks();

	GeCB gecb;
	gecb.signal_func = NULL;
	gecb.signal_arg = NULL;
	gecb.finish_func = Ge_Finish_Callback;
	gecb.finish_arg = NULL;
	gecbid = sceGeSetCallback(&gecb);

	load_menu_bg(PBPPath);
	strcpy(savedir,PBPPath);
	strcat(savedir,"SAVE");
	sceIoMkdir(savedir,0777);

	so.sound_fd = -1;

	memset( &Settings, 0, sizeof( Settings ) );

	Settings.CyclesPercentage = 100;
	Settings.SoundPlaybackRate = 4;
	Settings.SoundBufferSize = SOUND_SAMPLE;
	Settings.DisableSoundEcho = FALSE;
	Settings.APUEnabled = Settings.NextAPUEnabled = TRUE;
	Settings.SoundSync = FALSE;
	Settings.ThreadSound = TRUE;
	Settings.SoundEnvelopeHeightReading = TRUE;
	Settings.DisableMasterVolume = FALSE;
	Settings.ShutdownMaster = TRUE;
	Settings.SkipFrames = 0;
	Settings.FrameTimePAL = 20000;
	Settings.FrameTimeNTSC = 16667;
	Settings.FrameTime = Settings.FrameTimeNTSC;
	Settings.TurboMode = FALSE;
	Settings.TurboSkipFrames = 10;
	Settings.Transparency = TRUE;
	Settings.SupportHiRes = FALSE;
	Settings.ControllerOption = SNES_JOYPAD;
	Settings.H_Max = SNES_CYCLES_PER_SCANLINE;
	Settings.HBlankStart = (Settings.H_Max << 8) / SNES_HCOUNTER_MAX;
	Settings.PSP_APUTimerCycle = SNES_APUTIMER2_CYCLEx10000;

#ifndef OPTI
	Settings.JoystickEnabled = FALSE;
	Settings.Stereo = TRUE;
	Settings.StretchScreenshots = 1;
	Settings.Mouse = FALSE;
	Settings.SuperScope = FALSE;
	Settings.MultiPlayer5 = FALSE;
	Settings.SixteenBit = TRUE;
	Settings.NetPlay = FALSE;
	Settings.ServerName [0] = 0;
	Settings.AutoSaveDelay = 30;
	Settings.ApplyCheats = FALSE;
	Settings.StretchScreenshots = 1;
	Settings.Transparency = Settings.ForceTransparency;

	if ( Settings.ForceNoTransparency ){
		Settings.Transparency = FALSE;
	}

	if ( Settings.Transparency ){
		Settings.SixteenBit = TRUE;
	}
#endif // OPTI

	if ( !Memory.Init() || !S9xInitAPU() ){
		return -1;
	}

#ifdef OPTI
	S9xInitSound( Settings.SoundPlaybackRate, 1, Settings.SoundBufferSize );
#else
	S9xInitSound( Settings.SoundPlaybackRate, Settings.Stereo, Settings.SoundBufferSize );
#endif // OPTI

	if ( !Settings.APUEnabled ){
		S9xSetSoundMute( TRUE );
	}

	uint32 saved_flags = CPU.Flags;

	load_global_config();
	FilerMsg[0]=0;
	for(;;){
		while(!getFilePath(RomPath))
			;
		if ( !Memory.LoadROM(RomPath) ){
			continue;
		}
		S9xSetInfoString( "ROMC[Wǂݍ݂܂" );
		break;
	}
	pgFillvram(0);
	pgScreenFlipV();
	pgFillvram(0);
	pgScreenFlipV();
	Memory.LoadSRAM( S9xGetFilename("srm") );
	g_bROMLoaded = true;
	CPU.Flags = saved_flags;

	S9xInitInputDevices();

	S9xInitDisplay( 0, 0 );
	if ( !S9xGraphicsInit() ){
		return -1;
	}

	InitTimer();
	if ( !Settings.APUEnabled ){
		S9xSetSoundMute( FALSE );
	}

	s_iFrame = 0;
	sceKernelLibcGettimeofday(&s_tvStart, NULL);

	PSP_Settings.bSoundOff = false;
	load_config();

	S9xSetSoundMute( TRUE );

	while ( g_bLoop ){
		S9xMainLoop();
		
		if (g_bSleep){
			while(g_bSleep)
				pgWaitV();
			sceKernelLibcGettimeofday(&s_tvStart, NULL);
			next1.tv_sec = 0;
			if (PSP_Settings.iPSP_ClockUp == 2) scePowerSetClockFrequency(333,333,166);
			else if (PSP_Settings.iPSP_ClockUp == 1) scePowerSetClockFrequency(266,266,133);
			if (!PSP_Settings.bSoundOff) { S9xSetSoundMute( FALSE );}
		}
	}

	return 0;
}

}; // end of extern "C"

